/*  OpenTally: Open-source election vote counting
 *  Copyright © 2021 Lee Yingtong Li (RunasSudo)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

/// Fixed-point arithmetic (using `ibig`)
mod fixed;
/// Guarded fixed-point arithmetic (using `ibig`)
mod gfixed;
/// Native 64-bit floating point arithmetic
mod native;

/// Exact rational arithmetic (using `rug`/GMP)
#[cfg(not(target_arch = "wasm32"))]
mod rational_rug;

/// Exact rational arithmetic (using `num-bigint`)
//#[cfg(target_arch = "wasm32")]
mod rational_num;

//#[cfg(target_arch = "wasm32")]
mod dynnum;

use num_traits::{NumAssignRef, NumRef};

#[cfg(not(target_arch = "wasm32"))]
use rkyv::{Archive, Archived, Deserialize, Fallible, Resolver, Serialize, with::{ArchiveWith, DeserializeWith, SerializeWith}};

use std::cmp::Ord;
use std::fmt;
use std::ops;

/// Assign value, avoiding additional allocations
pub trait Assign<Src=Self> {
	/// Set the value of `self` to the value of `src`, avoiding additional allocations
	fn assign(&mut self, src: Src);
}

/// Trait for OpenTally numeric representations
//pub trait Number: NumRef + NumAssignRef + PartialOrd + Assign + Clone + fmt::Display where for<'a> &'a Self: RefNum<&'a Self> {
pub trait Number:
	NumRef + NumAssignRef + ops::Neg<Output=Self> + Ord + Assign + From<usize> + Clone + fmt::Debug + fmt::Display
where
	for<'a> Self: Assign<&'a Self>,
{
	/// Return a new [Number]
	fn new() -> Self;
	
	/// Convert to CLI argument representation
	fn describe() -> String;
	/// Convert to CLI argument representation, returning an empty string if the default
	fn describe_opt() -> String { Self::describe() }
	
	/// Exponentiate `self` to the `exponent` power
	fn pow_assign(&mut self, exponent: i32);
	/// Round `self` down if necessary to `dps` decimal places
	fn floor_mut(&mut self, dps: usize);
	/// Round `self` up if necessary to `dps` decimal places
	fn ceil_mut(&mut self, dps: usize);
	/// Round `self` half-up to the nearest `dps` decimal places
	fn round_mut(&mut self, dps: usize);
	
	/// Parse the given string into a [Number]
	fn parse(s: &str) -> Self {
		if let Ok(value) = Self::from_str_radix(s, 10) {
			return value;
		} else {
			panic!("Syntax Error");
		}
	}
}

/// rkyv-serialized representation of [Number]
#[cfg(not(target_arch = "wasm32"))]
pub struct SerializedNumber;

/// rkyv-serialized representation of [Option<Number>]
#[cfg(not(target_arch = "wasm32"))]
pub struct SerializedOptionNumber;

#[cfg(not(target_arch = "wasm32"))]
impl<N: Number> ArchiveWith<N> for SerializedNumber {
	type Archived = Archived<String>;
	type Resolver = Resolver<String>;
	
	unsafe fn resolve_with(field: &N, pos: usize, resolver: Self::Resolver, out: *mut Self::Archived) {
		field.to_string().resolve(pos, resolver, out);
	}
}

#[cfg(not(target_arch = "wasm32"))]
impl<N: Number> ArchiveWith<Option<N>> for SerializedOptionNumber {
	type Archived = Archived<String>;
	type Resolver = Resolver<String>;
	
	unsafe fn resolve_with(field: &Option<N>, pos: usize, resolver: Self::Resolver, out: *mut Self::Archived) {
		match field {
			Some(n) => {
				n.to_string().resolve(pos, resolver, out);
			}
			None => {
				String::new().resolve(pos, resolver, out);
			}
		}
	}
}

#[cfg(not(target_arch = "wasm32"))]
impl<N: Number, S: Fallible + ?Sized> SerializeWith<N, S> for SerializedNumber where String: Serialize<S> {
	fn serialize_with(field: &N, serializer: &mut S) -> Result<Self::Resolver, S::Error> {
		return field.to_string().serialize(serializer);
	}
}

#[cfg(not(target_arch = "wasm32"))]
impl<N: Number, S: Fallible + ?Sized> SerializeWith<Option<N>, S> for SerializedOptionNumber where String: Serialize<S> {
	fn serialize_with(field: &Option<N>, serializer: &mut S) -> Result<Self::Resolver, S::Error> {
		match field {
			Some(n) => {
				return n.to_string().serialize(serializer);
			}
			None => {
				return String::new().serialize(serializer);
			}
		}
	}
}

#[cfg(not(target_arch = "wasm32"))]
impl<N: Number, D: Fallible + ?Sized> DeserializeWith<Archived<String>, N, D> for SerializedNumber where Archived<String>: Deserialize<String, D> {
	fn deserialize_with(field: &Archived<String>, deserializer: &mut D) -> Result<N, D::Error> {
		return Ok(N::parse(&field.deserialize(deserializer)?));
	}
}

#[cfg(not(target_arch = "wasm32"))]
impl<N: Number, D: Fallible + ?Sized> DeserializeWith<Archived<String>, Option<N>, D> for SerializedOptionNumber where Archived<String>: Deserialize<String, D> {
	fn deserialize_with(field: &Archived<String>, deserializer: &mut D) -> Result<Option<N>, D::Error> {
		let s = field.deserialize(deserializer)?;
		if s.is_empty() {
			return Ok(None);
		} else {
			return Ok(Some(N::parse(&s)));
		}
	}
}
pub use self::dynnum::NumKind;
pub use self::dynnum::DynNum;

pub use self::fixed::Fixed;
pub use self::gfixed::GuardedFixed;
pub use self::native::NativeFloat64;

#[cfg(not(target_arch = "wasm32"))]
pub use self::rational_rug::Rational;

#[cfg(target_arch = "wasm32")]
pub use self::rational_num::Rational;
