/*  OpenTally: Open-source election vote counting
 *  Copyright © 2021 Lee Yingtong Li (RunasSudo)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use crate::utils;

use opentally::election::Election;
use opentally::numbers::Rational;
use opentally::parser::blt;
use opentally::stv;

use csv::StringRecord;
use flate2::bufread::GzDecoder;
use utf8_chars::BufReadCharsExt;

use std::fs::File;
use std::io::{self, BufReader};

#[test]
fn aec_tas19_rational() {
	// Read CSV file
	let reader = csv::ReaderBuilder::new()
		.has_headers(false)
		.from_path("tests/data/aec-senate-formalpreferences-24310-TAS.csv")
		.expect("IO Error");
	let records: Vec<StringRecord> = reader.into_records().map(|r| r.expect("Syntax Error")).collect();
	
	let mut candidates: Vec<&str> = records.iter().skip(2).map(|r| &r[0]).collect();
	// Remove exhausted/LBF rows
	candidates.truncate(candidates.len() - 2);
	
	let stages: Vec<usize> = records.first().unwrap().iter().skip(1).step_by(2).map(|s| s.parse().unwrap()).collect();
	
	// Decompress BLT
	let file = File::open("tests/data/aec-senate-formalpreferences-24310-TAS.blt.gz").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let gz_decoder = GzDecoder::new(file_reader);
	
	let mut reader = BufReader::new(gz_decoder);
	let chars = reader.chars().map(|r| r.expect("IO Error")).peekable();
	
	// Read BLT
	let election: Election<Rational> = blt::parse_iterator(chars).expect("Syntax Error");
	
	// Validate candidate names
	for (i, candidate) in candidates.iter().enumerate() {
		assert_eq!(election.candidates[i].name, *candidate);
	}
	
	let stv_opts = stv::STVOptionsBuilder::default()
		.round_votes(Some(0))
		.round_quota(Some(0))
		.quota_criterion(stv::QuotaCriterion::GreaterOrEqual)
		.surplus(stv::SurplusMethod::UIG)
		.surplus_order(stv::SurplusOrder::ByOrder)
		.exclusion(stv::ExclusionMethod::ByValue)
		.bulk_exclude(true)
		.build().unwrap();
	
	assert_eq!(stv_opts.describe::<Rational>(), "--round-votes 0 --round-quota 0 --quota-criterion geq --surplus uig --surplus-order by_order --exclusion by_value --bulk-exclude");
	
	utils::validate_election::<Rational>(stages, records, election, stv_opts, None, &["exhausted", "lbf"]);
}
