/*  OpenTally: Open-source election vote counting
 *  Copyright © 2021 Lee Yingtong Li (RunasSudo)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use crate::utils;

use opentally::constraints::Constraints;
use opentally::election::{CandidateState, CountState, Election};
use opentally::numbers::Rational;
use opentally::parser::blt;
use opentally::stv;

use std::fs::File;
use std::io::{self, BufRead}; 

#[test]
fn prsa1_constr1_rational() {
	// FIXME: This is unvalidated!
	
	// Read BLT
	let mut election: Election<Rational> = blt::parse_path("tests/data/prsa1.blt").expect("Syntax Error");
	
	// Read CON
	let file = File::open("tests/data/prsa1_constr1.con").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	election.constraints = Some(Constraints::from_con(lines.map(|r| r.expect("IO Error").to_string()).into_iter()).unwrap());
	
	let stv_opts = stv::STVOptionsBuilder::default()
		.round_surplus_fractions(Some(3))
		.round_values(Some(3))
		.round_votes(Some(3))
		.round_quota(Some(3))
		.quota_criterion(stv::QuotaCriterion::GreaterOrEqual)
		.surplus(stv::SurplusMethod::EG)
		.surplus_order(stv::SurplusOrder::ByOrder)
		.transferable_only(true)
		.exclusion(stv::ExclusionMethod::ParcelsByOrder)
		.early_bulk_elect(false)
		.constraints_path(Some("tests/data/prsa1_constr1.con".to_string()))
		.build().unwrap();
	
	// Initialise count state
	let mut state = CountState::new(&election);
	
	// Count election
	stv::count_init(&mut state, &stv_opts).unwrap();
	while stv::count_one_stage::<Rational>(&mut state, &stv_opts).unwrap() == false {}
	
	// Validate winners
	let mut winners = Vec::new();
	for (candidate, count_card) in state.candidates.iter() {
		if count_card.state == CandidateState::Elected {
			winners.push((candidate, count_card));
		}
	}
	winners.sort_unstable_by(|a, b| a.1.order_elected.partial_cmp(&b.1.order_elected).unwrap());
	
	assert_eq!(winners[0].0.name, "Grey");
	assert_eq!(winners[1].0.name, "Ames");
	assert_eq!(winners[2].0.name, "White");
	assert_eq!(winners[3].0.name, "Evans");
}

#[test]
fn prsa1_constr2_rational() {
	// FIXME: This is unvalidated!
	
	// Read BLT
	let mut election: Election<Rational> = blt::parse_path("tests/data/prsa1.blt").expect("Syntax Error");
	
	// Read CON
	let file = File::open("tests/data/prsa1_constr2.con").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	election.constraints = Some(Constraints::from_con(lines.map(|r| r.expect("IO Error").to_string()).into_iter()).unwrap());
	
	let stv_opts = stv::STVOptionsBuilder::default()
		.round_surplus_fractions(Some(3))
		.round_values(Some(3))
		.round_votes(Some(3))
		.round_quota(Some(3))
		.quota_criterion(stv::QuotaCriterion::GreaterOrEqual)
		.surplus(stv::SurplusMethod::EG)
		.surplus_order(stv::SurplusOrder::ByOrder)
		.transferable_only(true)
		.exclusion(stv::ExclusionMethod::ParcelsByOrder)
		.early_bulk_elect(false)
		.constraints_path(Some("tests/data/prsa1_constr1.con".to_string()))
		.build().unwrap();
	
	// Initialise count state
	let mut state = CountState::new(&election);
	
	// Count election
	stv::count_init(&mut state, &stv_opts).unwrap();
	while stv::count_one_stage::<Rational>(&mut state, &stv_opts).unwrap() == false {}
	
	// Validate winners
	let mut winners = Vec::new();
	for (candidate, count_card) in state.candidates.iter() {
		if count_card.state == CandidateState::Elected {
			winners.push((candidate, count_card));
		}
	}
	winners.sort_unstable_by(|a, b| a.1.order_elected.partial_cmp(&b.1.order_elected).unwrap());
	
	assert_eq!(winners[0].0.name, "Grey");
	assert_eq!(winners[1].0.name, "Ames");
	assert_eq!(winners[2].0.name, "White");
	assert_eq!(winners[3].0.name, "Evans");
}

#[test]
fn prsa1_constr3_rational() {
	// FIXME: This is unvalidated!
	
	// Read BLT
	let mut election: Election<Rational> = blt::parse_path("tests/data/prsa1.blt").expect("Syntax Error");
	
	// Read CON
	let file = File::open("tests/data/prsa1_constr3.con").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	election.constraints = Some(Constraints::from_con(lines.map(|r| r.expect("IO Error").to_string()).into_iter()).unwrap());
	
	let stv_opts = stv::STVOptionsBuilder::default()
		.round_surplus_fractions(Some(3))
		.round_values(Some(3))
		.round_votes(Some(3))
		.round_quota(Some(3))
		.quota_criterion(stv::QuotaCriterion::GreaterOrEqual)
		.surplus(stv::SurplusMethod::EG)
		.surplus_order(stv::SurplusOrder::ByOrder)
		.transferable_only(true)
		.exclusion(stv::ExclusionMethod::ParcelsByOrder)
		.early_bulk_elect(false)
		.constraints_path(Some("tests/data/prsa1_constr1.con".to_string()))
		.build().unwrap();
	
	// Initialise count state
	let mut state = CountState::new(&election);
	
	// Count election
	stv::count_init(&mut state, &stv_opts).unwrap();
	while stv::count_one_stage::<Rational>(&mut state, &stv_opts).unwrap() == false {}
	
	// Validate winners
	let mut winners = Vec::new();
	for (candidate, count_card) in state.candidates.iter() {
		if count_card.state == CandidateState::Elected {
			winners.push((candidate, count_card));
		}
	}
	winners.sort_unstable_by(|a, b| a.1.order_elected.partial_cmp(&b.1.order_elected).unwrap());
	
	assert_eq!(winners[0].0.name, "Grey");
	assert_eq!(winners[1].0.name, "White");
	assert_eq!(winners[2].0.name, "Thomson");
	assert_eq!(winners[3].0.name, "Reid");
}

/// Same election data as ers97_rational, but with a constraint that prevents the bulk exclusion of Glazier and Wright
#[test]
fn ers97old_cantbulkexclude_rational() {
	// Read CSV file
	let reader = csv::ReaderBuilder::new()
		.has_headers(false)
		.from_path("tests/data/ers97old_cantbulkexclude.csv")
		.expect("IO Error");
	let records: Vec<csv::StringRecord> = reader.into_records().map(|r| r.expect("Syntax Error")).collect();
	
	let mut candidates: Vec<&str> = records.iter().skip(2).map(|r| &r[0]).collect();
	// Remove exhausted/LBF rows
	candidates.truncate(candidates.len() - 2);
	
	let stages: Vec<usize> = records.first().unwrap().iter().skip(1).step_by(2).map(|s| s.parse().unwrap()).collect();
	
	// Read BLT
	let mut election: Election<Rational> = blt::parse_path("tests/data/ers97old.blt").expect("Syntax Error");
	
	// Read CON
	let file = File::open("tests/data/ers97old_cantbulkexclude.con").expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	election.constraints = Some(Constraints::from_con(lines.map(|r| r.expect("IO Error").to_string()).into_iter()).unwrap());
	
	let stv_opts = stv::STVOptionsBuilder::default()
		.round_surplus_fractions(Some(2))
		.round_values(Some(2))
		.round_votes(Some(2))
		.round_quota(Some(2))
		.quota(stv::QuotaType::DroopExact)
		.quota_criterion(stv::QuotaCriterion::GreaterOrEqual)
		.quota_mode(stv::QuotaMode::ERS97)
		.surplus(stv::SurplusMethod::EG)
		.transferable_only(true)
		.exclusion(stv::ExclusionMethod::ByValue)
		.early_bulk_elect(false)
		.bulk_exclude(true)
		.defer_surpluses(true)
		.constraints_path(Some("tests/data/ers97old_cantbulkexclude".to_string()))
		.build().unwrap();
	
	utils::validate_election::<Rational>(stages, records, election, stv_opts, None, &["nt", "vre"]);
}
